/*
 * IniEditor is Copyright (c) 2003-2005, Nik Haldimann
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *  
 * * Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *  
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package opt;

import java.io.*;
import java.util.*;

/**
 * <h1>IniEditor is Copyright (c) 2003-2005, Nik Haldimann</h1>
 * @author Nik Haldimann, me at ubique.ch
 * @version r4 (8/10/2005)
 * @revision $Id: IniEditor.java,v 1.25 2005/10/08 19:26:43 Nik Exp $
 */
public class IniEditor {

    private static boolean DEFAULT_CASE_SENSITIVITY = false;
    private Map<String, Section> sections;
    private List<String> sectionOrder;
    private String commonName;
    private char[] commentDelims;
    private boolean isCaseSensitive;
    private OptionFormat optionFormat;

    public IniEditor() { this(null, null); }
    public IniEditor(boolean isCaseSensitive) { this(null, null, isCaseSensitive); }
    public IniEditor(String commonName) { this(commonName, null); }
    public IniEditor(String commonName, boolean isCaseSensitive) { this(commonName, null, isCaseSensitive); }
    public IniEditor(char[] delims) { this(null, delims); }
    public IniEditor(char[] delims, boolean isCaseSensitive) { this(null, delims, isCaseSensitive); }
    public IniEditor(String commonName, char[] delims) { this(commonName, delims, DEFAULT_CASE_SENSITIVITY); }
    public IniEditor(String commonName, char[] delims, boolean isCaseSensitive) {
        this.sections = new HashMap<String, Section>();
        this.sectionOrder = new LinkedList<String>();
        this.isCaseSensitive = isCaseSensitive;
        if (commonName != null) {
            this.commonName = commonName;
            addSection(this.commonName);
        }
        this.commentDelims = delims;
        this.optionFormat = new OptionFormat(Section.DEFAULT_OPTION_FORMAT);
    }

    public void setOptionFormatString(String formatString) { this.optionFormat = new OptionFormat(formatString); }

    public String get(String section, String option) {
        if (hasSection(section)) {
            Section sect = getSection(section);
            if (sect.hasOption(option)) return sect.get(option);
            if (this.commonName != null) return getSection(this.commonName).get(option);
        }
        return null;
    }

    public void set(String section, String option, String value) {
        if (hasSection(section)) getSection(section).set(option, value);
        else throw new NoSuchSectionException(section);
    }

    public boolean remove(String section, String option) {
        if (hasSection(section)) return getSection(section).remove(option);
        else throw new NoSuchSectionException(section);
    }

    public boolean hasOption(String section, String option) { return hasSection(section) && getSection(section).hasOption(option); }
    public boolean hasSection(String name) { return this.sections.containsKey(normSection(name)); }

    public boolean addSection(String name) {
        String normName = normSection(name);
        if (!hasSection(normName)) {
            // Section constructor might throw IllegalArgumentException
            Section section = new Section(normName, this.commentDelims,
                                                         this.isCaseSensitive);
            section.setOptionFormat(this.optionFormat);
            this.sections.put(normName, section);
            this.sectionOrder.add(normName);
            return true;
        }
        else return false;
    }

    public boolean removeSection(String name) {
        String normName = normSection(name);
        if (this.commonName != null && this.commonName.equals(normName))
            throw new IllegalArgumentException("Can't remove common section");
        if (hasSection(normName)) {
            this.sections.remove(normName);
            this.sectionOrder.remove(normName);
            return true;
        }
        else return false;
    }

    public List sectionNames() {
        List<String> sectList = new ArrayList<String>(this.sectionOrder);
        if (this.commonName != null) sectList.remove(this.commonName);
        return sectList;
    }

    public List optionNames(String section) {
        if (hasSection(section)) return getSection(section).optionNames();
        else throw new NoSuchSectionException(section);
    }

    public void addComment(String section, String comment) {
        if (hasSection(section)) getSection(section).addComment(comment);
        else throw new NoSuchSectionException(section);
    }

    public void addBlankLine(String section) {
        if (hasSection(section)) getSection(section).addBlankLine();
        else throw new NoSuchSectionException(section);
    }

    public void save(String filename) throws IOException { save(new File(filename)); }
    public void save(File file) throws IOException {
        OutputStream out = new FileOutputStream(file);
        save(out);
        out.close();
    }
    public void save(OutputStream stream) throws IOException { save(new OutputStreamWriter(stream)); }
    public void save(OutputStreamWriter streamWriter) throws IOException {
        Iterator it = this.sectionOrder.iterator();
        PrintWriter writer = new PrintWriter(streamWriter, true);
        while (it.hasNext()) {
            Section sect = getSection((String)it.next());
            writer.println(sect.header());
            sect.save(writer);
        }
    }

    public void load(String filename) throws IOException { load(new File(filename)); }
    public void load(File file) throws IOException {
        InputStream in = new FileInputStream(file);
        load(in);
        in.close();
    }
    public void load(InputStream stream) throws IOException { load(new InputStreamReader(stream)); }
    public void load(InputStreamReader streamReader) throws IOException {
        BufferedReader reader = new BufferedReader(streamReader);
        String curSection = null;
        String line = null;

        while (reader.ready()) {
            line = reader.readLine().trim();
            if (line.length() > 0 && line.charAt(0) == Section.HEADER_START) {
                int endIndex = line.indexOf(Section.HEADER_END);
                if (endIndex >= 0) {
                    curSection = line.substring(1, endIndex);
                    addSection(curSection);
                }
            }
            if (curSection != null) {
                Section sect = getSection(curSection);
                sect.load(reader);
            }
        }
    }

    private Section getSection(String name) { return (Section)sections.get(normSection(name)); }

    private String normSection(String name) {
        if (!this.isCaseSensitive) name = name.toLowerCase();
        return name.trim();
    }

    public static String[] toStringArray(Collection coll) {
        Object[] objArray = coll.toArray();
        String[] strArray = new String[objArray.length];
        for (int i = 0; i < objArray.length; i++) {
            strArray[i] = (String)objArray[i];
        }
        return strArray;
    }

    public static class Section {
        private String name;
        private Map<String, Option> options;
        private List<Line> lines;
        private char[] optionDelims;
        private char[] optionDelimsSorted;
        private char[] commentDelims;
        private char[] commentDelimsSorted;
        private boolean isCaseSensitive;
        private OptionFormat optionFormat;
        private static final char[] DEFAULT_OPTION_DELIMS = new char[] {'=', ':'};
        private static final char[] DEFAULT_COMMENT_DELIMS = new char[] {'#', ';'};
        private static final char[] OPTION_DELIMS_WHITESPACE = new char[] {' ', '\t'};
        private static final boolean DEFAULT_CASE_SENSITIVITY = false;
        public static final String DEFAULT_OPTION_FORMAT = "%s %s %s";
        public static final char HEADER_START = '[';
        public static final char HEADER_END = ']';
        private static final int NAME_MAXLENGTH = 1024;
        private static final char[] INVALID_NAME_CHARS = {HEADER_START, HEADER_END};

        public Section(String name) { this(name, null); }
        public Section(String name, boolean isCaseSensitive) { this(name, null, isCaseSensitive); }
        public Section(String name, char[] delims) { this(name, delims, DEFAULT_CASE_SENSITIVITY); }
        public Section(String name, char[] delims, boolean isCaseSensitive) {
            if (!validName(name))  throw new IllegalArgumentException("Illegal section name:" + name);
            this.name = name;
            this.isCaseSensitive = isCaseSensitive;
            this.options = new HashMap<String, Option>();
            this.lines = new LinkedList<Line>();
            this.optionDelims = DEFAULT_OPTION_DELIMS;
            this.commentDelims = (delims == null ? DEFAULT_COMMENT_DELIMS : delims);
            this.optionFormat = new OptionFormat(DEFAULT_OPTION_FORMAT);
            this.optionDelimsSorted = new char[this.optionDelims.length];
            System.arraycopy(this.optionDelims, 0, this.optionDelimsSorted, 0, this.optionDelims.length);
            this.commentDelimsSorted = new char[this.commentDelims.length];
            System.arraycopy(this.commentDelims, 0, this.commentDelimsSorted, 0, this.commentDelims.length);
            Arrays.sort(this.optionDelimsSorted);
            Arrays.sort(this.commentDelimsSorted);
        }

        public void setOptionFormatString(String formatString) { this.setOptionFormat(new OptionFormat(formatString)); }
        public void setOptionFormat(OptionFormat format) { this.optionFormat = format; }

        public List optionNames() {
            List<String> optNames = new LinkedList<String>();
            Iterator it = this.lines.iterator();
            while (it.hasNext()) {
                Object line = it.next();
                if (line instanceof Option) optNames.add(((Option)line).name());
            }
            return optNames;
        }

        public boolean hasOption(String name) { return this.options.containsKey(normOption(name)); }

        public String get(String option) {
            String normed = normOption(option);
            if (hasOption(normed)) return getOption(normed).value();
            return null;
        }

        public void set(String option, String value) { set(option, value, this.optionDelims[0]); }
        public void set(String option, String value, char delim) {
            String normed = normOption(option);
            if (hasOption(normed)) getOption(normed).set(value);
            else {
                Option opt = new Option(normed, value, delim, this.optionFormat);
                this.options.put(normed, opt);
                this.lines.add(opt);
            }
        }

        public boolean remove(String option) {
            String normed = normOption(option);
            if (hasOption(normed)) {
                this.lines.remove(getOption(normed));
                this.options.remove(normed);
                return true;
            }
            else return false;
        }

        public void addComment(String comment) { addComment(comment, this.commentDelims[0]); }
        public void addComment(String comment, char delim) {
            StringTokenizer st = new StringTokenizer(comment.trim(), NEWLINE_CHARS);
            while (st.hasMoreTokens()) this.lines.add(new Comment(st.nextToken(), delim));
        }
        private static final String NEWLINE_CHARS = "\n\r";

        public void addBlankLine() { this.lines.add(BLANK_LINE); }

        public void load(BufferedReader reader) throws IOException {
            while (reader.ready()) {
                reader.mark(NAME_MAXLENGTH);
                String line = reader.readLine().trim();

                if (line.length() > 0 && line.charAt(0) == HEADER_START) {
                    reader.reset();
                    return;
                }

                int delimIndex = -1;
                if (line.equals("")) this.addBlankLine();
                else if ((delimIndex = Arrays.binarySearch(this.commentDelimsSorted, line.charAt(0))) >= 0) 
                    addComment(line.substring(1), this.commentDelimsSorted[delimIndex]);
                else {
                    delimIndex = -1;
                    int delimNum = -1;
                    int lastSpaceIndex = -1;
                    for (int i = 0, l = line.length(); i < l && delimIndex < 0; i++) {
                        delimNum = Arrays.binarySearch(this.optionDelimsSorted, line.charAt(i));
                        if (delimNum >= 0) delimIndex = i;
                        else {
                            boolean isSpace = Arrays.binarySearch(OPTION_DELIMS_WHITESPACE, line.charAt(i)) >= 0;
                            if(!isSpace && lastSpaceIndex >= 0) break;
                            else if (isSpace) lastSpaceIndex = i;
                        }
                    }
                    if (delimIndex < 0) {
                        if (lastSpaceIndex < 0) this.set(line, "");
                        else this.set(line.substring(0, lastSpaceIndex), line.substring(lastSpaceIndex+1));
                    }
                    else this.set(line.substring(0, delimIndex), line.substring(delimIndex+1), line.charAt(delimIndex));
                }
            }
        }

        public void save(PrintWriter writer) throws IOException {
            Iterator it = this.lines.iterator();
            while (it.hasNext())  writer.println(((Line)it.next()).toString());
            if (writer.checkError()) throw new IOException();
        }

        private Option getOption(String name) { return (Option)this.options.get(name); }
        private String header() { return HEADER_START + this.name + HEADER_END; }

        private static boolean validName(String name) {
            if (name.trim().equals("")) return false;
            for (int i = 0; i < INVALID_NAME_CHARS.length; i++)
                if (name.indexOf(INVALID_NAME_CHARS[i]) >= 0) return false;
            return true;
        }

        private String normOption(String name) {
            if (!this.isCaseSensitive) name = name.toLowerCase();
            return name.trim();
        }
    }

    private interface Line { public String toString(); }
    private static final Line BLANK_LINE = new Line() { public String toString() { return ""; } };

    private static class Option implements Line {
        private String name;
        private String value;
        private char separator;
        private OptionFormat format;
        private static final String ILLEGAL_VALUE_CHARS = "\n\r";

        public Option(String name, String value, char separator, OptionFormat format) {
            if (!validName(name, separator)) throw new IllegalArgumentException("Illegal option name:" + name);
            this.name = name;
            this.separator = separator;
            this.format = format;
            set(value);
        }

        public String name() { return this.name; }
        public String value() { return this.value; }
        public void set(String value) {
            if (value == null) this.value = value;
            else {
                StringTokenizer st = new StringTokenizer(value.trim(), ILLEGAL_VALUE_CHARS);
                StringBuffer sb = new StringBuffer();
                while (st.hasMoreTokens()) sb.append(st.nextToken());
                this.value = sb.toString();
            }
        }

        public String toString() { return this.format.format(this.name, this.value, this.separator); }
        private static boolean validName(String name, char separator) {
            if (name.trim().equals("")) return false;
            if (name.indexOf(separator) >= 0) return false;
            return true;
        }
    }

    private static class Comment implements Line {
        private String comment;
        private char delimiter;
        private static final char DEFAULT_DELIMITER = '#';
        public Comment(String comment) { this(comment, DEFAULT_DELIMITER); }
        public Comment(String comment, char delimiter) {
            this.comment = comment.trim();
            this.delimiter = delimiter;
        }
        public String toString() { return this.delimiter + " " + this.comment; }
    }

    private static class OptionFormat {
        private static final int EXPECTED_TOKENS = 4;
        private String[] formatTokens;
        public OptionFormat(String formatString) { this.formatTokens = this.compileFormat(formatString); }
        public String format(String name, String value, char separator) {
            String[] t = this.formatTokens;
            return t[0] + name + t[1] + separator + t[2] + value + t[3];
        }
        
        private String[] compileFormat(String formatString) {
            String[] tokens = {"", "", "", ""};
            int tokenCount = 0;
            boolean seenPercent = false;
            StringBuffer token = new StringBuffer();
            for (int i = 0; i < formatString.length(); i++) {
                switch (formatString.charAt(i)) {
                    case '%':
                        if (seenPercent) {
                            token.append("%");
                            seenPercent = false;
                        }
                        else seenPercent = true;
                        break;
                    case 's':
                        if (seenPercent) {
                            if (tokenCount >= EXPECTED_TOKENS) throw new IllegalArgumentException("Illegal option format. Too many %s placeholders.");
                            tokens[tokenCount] = token.toString();
                            tokenCount++;
                            token = new StringBuffer();
                            seenPercent = false;
                        }
                        else token.append("s");
                        break;
                    default:
                        if (seenPercent) throw new IllegalArgumentException("Illegal option format. Unknown format specifier.");
                        token.append(formatString.charAt(i));
                        break;
                }
            }
            if (tokenCount != EXPECTED_TOKENS - 1) throw new IllegalArgumentException("Illegal option format. Not enough %s placeholders.");
            tokens[tokenCount] = token.toString();
            return tokens;
        }
    }

    public static class NoSuchSectionException extends RuntimeException {
		private static final long serialVersionUID = 8400415011278438260L;
		public NoSuchSectionException() { super(); }
        public NoSuchSectionException(String msg) { super(msg); }
    }
}
